# Taken from https://github.com/softwarecrafts/django-cptemplate
#
# MIT License
#
# Copyright (c) 2024 Andrew Miller
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

from pathlib import Path
from shutil import copy2

from django.apps import apps
from django.conf import settings
from django.core.management.base import BaseCommand
from django.core.management.base import CommandError
from django.template import loader
from django.template.exceptions import TemplateDoesNotExist
from django.template.exceptions import TemplateSyntaxError


def get_template_absolute_path(template_path):
    try:
        template = loader.get_template(template_path)
        return template.origin.name  # type: ignore[attr-defined]
    except TemplateDoesNotExist:
        msg = f"Template {template_path} was not found"
        raise CommandError(msg)
    except TemplateSyntaxError as e:
        msg = f"Syntax error in template {template_path}: {e}"
        raise CommandError(msg) from e


class Command(BaseCommand):
    help = "Copies a template from a package into your project, if no destination is provided, it will copy the template to the project level templates directory"

    def add_arguments(self, parser):
        parser.add_argument(
            "source",
            type=str,
            help="The path to the template to copy, e.g. 'account/login.html' for the allauth login template",
        )
        parser.add_argument(
            "destination",
            type=str,
            nargs="?",
            help="An optional app where the template should be copied",
        )

    def handle(self, *args, **options):
        # 1. extract the options
        source = options["source"]
        destination = options.get("destination")
        # 2. Check that the source file exists - use template loaders for this
        source_file = get_template_absolute_path(source)
        base_dir = Path(settings.BASE_DIR)  # type: ignore[misc]
        # 4. if destination, then create Path object and copy
        if destination is not None:
            app_config = apps.get_app_config(destination)
            destination_path = base_dir / app_config.path / "templates" / source
        # 5/ else inspect TEMPLATES[DIRS] setting and use first option if available
        else:
            try:
                template_dir = Path(settings.TEMPLATES[0]["DIRS"][0])
                destination_path = template_dir / source
            except IndexError:
                # 6/ otherwise create project level template directory and dump file there.
                destination_path = base_dir / "templates" / source
                # PRINT what settings need to be modified
                self.stdout.write(
                    self.style.WARNING(
                        'Update TEMPLATES["DIRS"] to include the following entry "BASE_DIR / "templates","'
                    )
                )

        destination_path.parent.mkdir(parents=True, exist_ok=True)
        copy2(source_file, destination_path)



import subprocess
from pathlib import Path

from django.conf import settings
from django.core.management.base import BaseCommand
from django.core.management.base import CommandError


def get_apps_dir() -> Path:
    try:
        return settings.APPS_DIR
    except AttributeError:
        msg = "Add an APPS_DIR settings, eg: APPS_DIR = BASE / 'apps'"
        raise CommandError(msg)


def exit_if_debug_false():
    if not settings.DEBUG:
        msg = "Nope, not happening, this command can only be run with DEBUG=True."
        raise CommandError(msg)


class CleanRepoOnlyCommand(BaseCommand):
    def add_arguments(self, parser):
        parser.add_argument("--allow-dirty", action="store_true", help="Allow dirty git repo.")
        super().add_arguments(parser)

    def handle(self, *args, **options):
        allow_dirty = options["allow_dirty"]
        result = subprocess.run(["git", "status", "--porcelain"], capture_output=True, text=True, check=False)
        git_repo_clean = result.stdout.strip() == ""
        if not git_repo_clean and not allow_dirty:
            msg = "Git repo is not clean, clean or stash away changes before running this command"
            raise CommandError(msg)
        return super().handle(*args, **options)



import importlib.util
import subprocess
from contextlib import contextmanager

from rich.progress import Progress
from rich.progress import SpinnerColumn
from rich.progress import TextColumn

if importlib.util.find_spec("falco_cli"):
    from falco_cli.utils import run_html_formatters, run_python_formatters
else:

    def run_html_formatters(*_, **__):
        pass

    def run_python_formatters(*_, **__):
        pass


@contextmanager
def simple_progress(description: str, display_text="[progress.description]{task.description}"):
    progress = Progress(SpinnerColumn(), TextColumn(display_text), transient=True)
    progress.add_task(description=description, total=None)
    try:
        yield progress.start()
    finally:
        progress.stop()





from importlib.metadata import version as get_version

# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

import django

import os, sys
os.environ["DJANGO_SETTINGS_MODULE"] = "demo.demo.settings"
sys.path.insert(0, os.path.abspath('..'))
from django.conf import settings
settings.configure(INSTALLED_APPS = ["falco"])
django.setup()

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "Falco"
copyright = "Copyright &copy; 2023, Tobi DEGNON"
author = "Tobi DEGNON"
version = get_version("falco-app")
release = version

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.extlinks",
    "myst_parser",
    "sphinx.ext.todo",
    "sphinx.ext.autodoc",
    "sphinx_design",
    "sphinx_tabs.tabs",
    "sphinx_copybutton",
    "sphinx_togglebutton",
    "sphinxcontrib.mermaid",
    # "sphinx_docsearch",
    "sphinx_exec_code",
]
todo_include_todos = True
extlinks = {
    "pull": ("https://github.com/falcopackages/falco-app/pull/%s", "pull request #%s"),
    "issue": ("https://github.com/falcopackages/falco-app/issues/%s", "issue #%s"),
    "repo": ("https://github.com/falcopackages/falco-app", "github repository"),
}

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "shibuya"
html_static_path = ["_static"]
html_baseurl = "https://falco.oluwatobi.dev"
html_title = "Falco"

# -- Shibuya theme options ---------------------------------------------------
html_context = {
    "source_type": "github",
    "source_user": "falcopackages",
    "source_repo": "falco-app",
}
html_theme_options = {
    "mastodon_url": "https://fosstodon.org/@tobide",
    "github_url": "https://github.com/falcopackages/falco-app",
    "twitter_url": "https://twitter.com/tobidegnon",
    "discussion_url": "https://github.com/falcopackages/falco-app/discussions",
    "accent_color": "blue",
    "globaltoc_expand_depth": 1,
}
html_logo = "https://raw.githubusercontent.com/falcopackages/website/refs/heads/main/_static/logo_with_text.svg"
html_favicon = "https://raw.githubusercohtml_logontent.com/falcopackages/website/refs/heads/main/_static/falco-logo.svg"

# html_sidebars = {
#     "**": [
#         "sidebars/localtoc.html",
#         "sidebars/repo-stats.html",
#         "sidebars/edit-this-page.html",
#         "sidebars/consulting.html",
#         # "sidebars/buy-me-a-coffee.html",
#     ]
# }

# html_js_files = [
#     (
#         "https://plausible.service.dotfm.me/js/script.js",
#         {"defer": "", "data-domain": "falco.oluwatobi.dev"},
#     ),
#     "add-og-title-to-home.js",
#     (
#         "https://cdnjs.buymeacoffee.com/1.0.0/widget.prod.min.js",
#         {
#             "data-name": "BMC-Widget",
#             "data-cfasync": "false",
#             "data-id": "oluwa.tobi",
#             "data-description": "Support me on Buy me a coffee!",
#             "data-message": "",
#             "data-color": "#5F7FFF",
#             "data-position": "Right",
#             "data-x_margin": "18",
#             "data-y_margin": "18",
#         },
#     ),
# ]
#
# # Jupyter sphinx configs
# jupyter_sphinx_thebelab_config = {
#     "requestKernel": True,
# }
# jupyter_sphinx_require_url = ""

# -- DocSearch configs -----------------------------------------------------
# docsearch_app_id = "CJEHOB5X2Y"
# docsearch_api_key = "e467f62765922e10749dec55f81a0a76"
# docsearch_index_name = "falco"



from django.core.paginator import InvalidPage
from django.core.paginator import Paginator
from django.db.models import QuerySet
from django.http import Http404

from falco.conf import app_settings
from falco.types import HttpRequest


def paginate_queryset(request: HttpRequest, queryset: QuerySet, page_size: int = app_settings.DEFAULT_PAGE_SIZE):
    paginator = Paginator(queryset, page_size)
    page_number = request.GET.get("page") or 1
    try:
        page_number = int(page_number)
    except ValueError as e:
        if page_number == "last":
            page_number = paginator.num_pages
        else:
            msg = "Page is not 'last', nor can it be converted to an int."
            raise Http404(msg) from e

    try:
        return paginator.page(page_number)
    except InvalidPage as exc:
        msg = "Invalid page (%s): %s"
        raise Http404(msg % (page_number, str(exc))) from exc



from __future__ import annotations

import logging

import pytest
from django.http import HttpRequest

pytest_plugins = []


def pytest_configure(config):
    logging.disable(logging.CRITICAL)


@pytest.fixture
def req():
    return HttpRequest()



from pathlib import Path

import httpx
from django.core.management.base import BaseCommand
from django.core.management.base import CommandError
from falco.management.commands.htmx import default_htmx_output_folder
from falco.management.commands.htmx import network_request_with_progress
from falco.utils import simple_progress
from rich import print as rich_print
from rich.console import Console
from rich.panel import Panel
from rich.table import Table

REGISTRY_URL = "https://htmx-extensions.oluwatobi.dev/extensions.json"


class Command(BaseCommand):
    help = "Download one of htmx extensions"

    def add_arguments(self, parser):
        parser.add_argument(
            "name",
            type=str,
            nargs="?",
            help="The name of the extension to download, if not specified will list all extensions",
            default=None,
        )
        parser.add_argument(
            "-o",
            "--output-dir",
            type=Path,
            default=None,
            help="The directory to write the downloaded file to.",
        )

    def handle(self, *_, **options):
        output_dir = options.get("output_dir") or default_htmx_output_folder()
        name = options.get("name")
        if name:
            self.download(name, output_dir=output_dir)
        else:
            self.list_all()

    def download(self, name: str, output_dir: Path):
        extensions = self.read_registry()
        extension = extensions.get(name)

        if not extension:
            msg = f"Could not find {name} extension."
            raise CommandError(msg)

        with simple_progress(f"Downloading {name} extension"):
            download_url = extension.get("download_url")
            response = httpx.get(download_url, follow_redirects=True)

            output_file = output_dir / f"{name}.js"
            output_file.write_text(response.text)

        rich_print(
            Panel(
                f"[green]Extension {name} downloaded successfully![/green]",
                subtitle=extension.get("doc_url"),
            )
        )

    def list_all(self):
        extensions = self.read_registry()
        table = Table(
            title="Htmx Extensions",
            caption="Full details at https://htmx-extensions.oluwatobi.dev",
            show_lines=True,
        )
        table.add_column("Name", style="green")
        table.add_column("Description", style="magenta")

        for name, metadata in extensions.items():
            table.add_row(name, metadata.get("description", ""))
        console = Console()
        console.print(table)

    @classmethod
    def read_registry(cls):
        with network_request_with_progress(REGISTRY_URL, "Loading extensions registry") as response:
            return response.json()



from django.contrib.auth.models import User
from django.db import models
from falco_toolbox.models import TimeStamped


class Entry(TimeStamped):
    # the TimeStampedModel adds the fields `created` and `modified` so we don't need to add them
    title = models.CharField(max_length=255)
    content = models.TextField()
    created_by = models.ForeignKey(User, on_delete=models.CASCADE, related_name="entries")

    class Meta:
        verbose_name = "Entry"
        verbose_name_plural = "Entries"



from django.db import models
from django.utils.text import slugify

from falco.models import TimeStamped


class Book(TimeStamped):
    lookup_field = "slug"

    name = models.CharField(max_length=255)
    slug = models.SlugField()
    description = models.TextField()
    published_at = models.DateField()
    on_going = models.BooleanField(default=True)
    cover_art = models.FileField(upload_to="covers", blank=True)
    author = models.ForeignKey("books.Author", on_delete=models.CASCADE)

    def __str__(self):
        return self.name

    def save(self, *args, **kwargs):
        if self._state.adding and not self.slug:
            self.slug = slugify(self.name)
        return super().save(*args, **kwargs)


class Author(TimeStamped):
    name = models.CharField(max_length=255)

    def __str__(self):
        return self.name



import pytest
from falco.management.commands.work import default_address
from falco.management.commands.work import default_server_cmd
from falco.management.commands.work import Work
from falco_cli.config import write_falco_config


def test_env_resolution(tmp_path):
    assert Work().resolve_django_env()


def test_env_resolution_with_env(tmp_path):
    (tmp_path / ".env").write_text("FOO=BAR")
    assert "FOO" in Work().resolve_django_env()


def test_without_pyproject_file():
    assert Work().get_commands() == {"server": default_server_cmd.format(address=default_address)}


def test_with_pyproject_file(pyproject_toml):
    write_falco_config(pyproject_path=pyproject_toml, work={"qcluster": "python manage.py qcluster"})
    assert Work().get_commands() == {
        "server": default_server_cmd.format(address=default_address),
        "qcluster": "python manage.py qcluster",
    }


def test_override_server(pyproject_toml):
    work = {
        "server": "python manage.py runserver",
        "qcluster": "python manage.py qcluster",
    }
    write_falco_config(pyproject_path=pyproject_toml, work=work)
    assert Work().get_commands() == work


@pytest.mark.parametrize("address", ["8000", "localhost:8000", "8001", "127.0.0.1"])
def test_override_server_through_arg(address):
    assert Work(address=address).get_commands() == {"server": default_server_cmd.format(address=address)}


@pytest.mark.parametrize("address", ["8000", "localhost:8000", "8001", "127.0.0.1"])
def test_override_server_through_arg_by_pyproject(pyproject_toml, address):
    work = {"server": "python manage.py runserver {address}"}
    write_falco_config(pyproject_path=pyproject_toml, work=work)
    assert Work(address=address).get_commands() == {"server": work["server"].format(address=address)}



import subprocess
from pathlib import Path

import cappa
import pytest
from cappa.testing import CommandRunner


def makemigaration():
    subprocess.run(["python", "manage.py", "makemigrations"], check=False)


def test_rm_migrations(django_project, runner: CommandRunner, set_git_repo_to_clean):
    apps_dir = Path()
    makemigaration()
    first_migration = apps_dir / "blog/migrations/0001_initial.py"
    assert first_migration.exists()
    runner.invoke("rm-migrations", ".")
    assert not first_migration.exists()


def test_rm_migrations_fake_apps_dir(django_project, runner: CommandRunner, set_git_repo_to_clean):
    apps_dir = Path()
    makemigaration()
    first_migration = apps_dir / "blog/migrations/0001_initial.py"
    assert first_migration.exists()
    runner.invoke("rm-migrations", "myproject")
    assert first_migration.exists()


def test_rm_migrations_not_clean_repo(django_project, runner: CommandRunner):
    with pytest.raises(cappa.Exit):
        runner.invoke("rm-migrations", ".")



from typing import TypeVar

from django import template
from django.contrib.auth.models import AbstractUser
from django.contrib.auth.models import AnonymousUser
from django.db import models

register = template.Library()

_TModel = TypeVar("_TModel", bound=models.Model)


@register.filter()
def lookup(obj: _TModel) -> str:
    lookup_field = getattr(obj, "lookup_field", "pk")
    return str(getattr(obj, lookup_field))


@register.filter()
def field_verbose_names(objects, fields) -> list[str]:
    return [objects[0]._meta.get_field(f).verbose_name for f in fields]  # noqa


@register.filter(name="getattr")
def get_attribute(obj: object, field: str):
    return getattr(obj, field)


@register.filter()
def field_class_name(obj: _TModel, field: str) -> str:
    return obj._meta.get_field(field).__class__.__name__  # noqa


@register.filter
def display_name(user: AbstractUser) -> str:
    if not hasattr(user, "username") or isinstance(user, AnonymousUser):
        return "N/A"

    if user.first_name and user.last_name:
        display_name_ = f"{user.first_name} {user.last_name}"
    elif user.first_name:
        display_name_ = user.first_name
    else:
        display_name_ = user.username

    return display_name_.title()


@register.filter
def initials(user: AbstractUser) -> str:
    if not hasattr(user, "username") or isinstance(user, AnonymousUser):
        return "N/A"

    initials_ = user.username[0].upper()

    if user.first_name and user.last_name:
        initials_ = f"{user.first_name[0].upper()}{user.last_name[0].upper()}"

    return initials_


@register.filter
def class_name(instance: object) -> str:
    return instance.__class__.__name__


@register.filter
def call_get_display(obj, field_name):
    method_name = f"get_{field_name}_display"
    method = getattr(obj, method_name, None)
    if callable(method):
        return method()



import subprocess
from pathlib import Path

import cappa
import pytest
from cappa.testing import CommandRunner
from falco_cli.config import write_falco_config

views_functions = ["post_list", "post_detail", "post_update", "post_create"]
html_templates = [
    "post_list.html",
    "post_detail.html",
    "post_update.html",
    "post_create.html",
]
views_functions_entry_point = ["index", "detail", "update", "create"]
html_templates_point = [
    "index.html",
    "detail.html",
    "update.html",
    "create.html",
]
forms_attributes = ["PostForm", "Post", "title", "content"]
admin_attributes = ["PostAdmin", "Post", "title", "content"]


def create_pyproject_crud_config(**kwargs):
    pyproject_toml = Path("pyproject.toml")
    pyproject_toml.touch()
    write_falco_config(pyproject_path=pyproject_toml, crud=kwargs)


def healthy_django_project() -> bool:
    result = subprocess.run(
        ["python", "manage.py", "check"],
        check=False,
        capture_output=True,
        text=True,
    )
    return result.returncode == 0


def test_crud(django_project, runner: CommandRunner, set_git_repo_to_clean):
    runner.invoke("crud", "blog.post")
    assert healthy_django_project()
    app_dir = Path("blog")
    assert (app_dir / "urls.py").exists()
    # sourcery skip: no-loop-in-tests
    for a in forms_attributes:
        assert a in (app_dir / "forms.py").read_text()

    for a in admin_attributes:
        assert a in (app_dir / "admin.py").read_text()

    for f in views_functions:
        assert f in (app_dir / "views.py").read_text()
    for t in html_templates:
        assert (app_dir / "templates" / "blog" / f"{t}").exists()


def test_crud_all_models(django_project, runner: CommandRunner, set_git_repo_to_clean):
    runner.invoke("crud", "blog")
    assert healthy_django_project()
    app_dir = Path("blog")
    assert (app_dir / "urls.py").exists()
    # sourcery skip: no-loop-in-tests
    for a in forms_attributes:
        assert a in (app_dir / "forms.py").read_text()

    for a in admin_attributes:
        assert a in (app_dir / "admin.py").read_text()

    for f in views_functions:
        assert f in (app_dir / "views.py").read_text()
    for t in html_templates:
        assert (app_dir / "templates" / "blog" / f"{t}").exists()


def test_crud_login(django_project, runner: CommandRunner, set_git_repo_to_clean):
    runner.invoke("crud", "blog.post", "--login-required")
    assert healthy_django_project()
    app_dir = Path("blog")
    assert (app_dir / "urls.py").exists()
    for a in forms_attributes:
        assert a in (app_dir / "forms.py").read_text()

    for a in admin_attributes:
        assert a in (app_dir / "admin.py").read_text()

    # sourcery skip: no-loop-in-tests
    for f in views_functions:
        assert f in (app_dir / "views.py").read_text()
    for t in html_templates:
        assert (app_dir / "templates" / "blog" / f"{t}").exists()


def test_crud_entry_point(django_project, runner: CommandRunner, set_git_repo_to_clean):
    runner.invoke("crud", "blog.post", "--entry-point")
    assert healthy_django_project()
    app_dir = Path("blog")
    assert (app_dir / "urls.py").exists()
    for a in forms_attributes:
        assert a in (app_dir / "forms.py").read_text()

    for a in admin_attributes:
        assert a in (app_dir / "admin.py").read_text()
    # sourcery skip: no-loop-in-tests
    for f in views_functions_entry_point:
        assert f in (app_dir / "views.py").read_text()
    for t in html_templates_point:
        assert (app_dir / "templates" / "blog" / f"{t}").exists()


def test_crud_entry_point_login(django_project, runner: CommandRunner, set_git_repo_to_clean):
    runner.invoke("crud", "blog.post", "--entry-point", "--login-required")
    assert healthy_django_project()
    app_dir = Path("blog")
    assert (app_dir / "urls.py").exists()
    for a in forms_attributes:
        assert a in (app_dir / "forms.py").read_text()

    for a in admin_attributes:
        assert a in (app_dir / "admin.py").read_text()
    # sourcery skip: no-loop-in-tests
    for f in views_functions_entry_point:
        assert f in (app_dir / "views.py").read_text()
    for t in html_templates_point:
        assert (app_dir / "templates" / "blog" / f"{t}").exists()


def test_crud_only_html(django_project, runner: CommandRunner, set_git_repo_to_clean):
    runner.invoke("crud", "blog.post", "--only-html")
    assert healthy_django_project()
    app_dir = Path("blog")
    assert not (app_dir / "urls.py").exists()
    assert not (app_dir / "forms.py").exists()
    # sourcery skip: no-loop-in-tests
    for f in views_functions:
        assert f not in (app_dir / "views.py").read_text()
    for t in html_templates:
        assert (app_dir / "templates" / "blog" / f"{t}").exists()


def test_crud_only_python(django_project, runner: CommandRunner, set_git_repo_to_clean):
    runner.invoke("crud", "blog.post", "--only-python")
    assert healthy_django_project()
    app_dir = Path("blog")
    assert (app_dir / "urls.py").exists()
    # sourcery skip: no-loop-in-tests
    for a in forms_attributes:
        assert a in (app_dir / "forms.py").read_text()

    for a in admin_attributes:
        assert a in (app_dir / "admin.py").read_text()
    # sourcery skip: no-loop-in-tests
    for f in views_functions:
        assert f in (app_dir / "views.py").read_text()
    for t in html_templates:
        assert not (app_dir / "templates" / "blog" / f"{t}").exists()


def test_crud_repo_not_clean(django_project, runner: CommandRunner):
    with pytest.raises(cappa.Exit):
        runner.invoke("crud", "blog.post")


def test_crud_exclude_field(django_project, runner: CommandRunner, set_git_repo_to_clean):
    runner.invoke("crud", "blog.post", "-e=title")
    app_dir = Path("blog")
    # sourcery skip: no-loop-in-tests
    assert "title" not in (app_dir / "forms.py").read_text()
    assert "title" not in (app_dir / "views.py").read_text()
    assert "post.title" not in (app_dir / "templates/blog/post_list.html").read_text()
    forms_attributes_ = ["PostForm", "Post", "content"]
    for a in forms_attributes_:
        assert a in (app_dir / "forms.py").read_text()


def test_crud_login_required(django_project, runner: CommandRunner, set_git_repo_to_clean):
    runner.invoke("crud", "blog.post", "--login-required")
    assert healthy_django_project()
    views = (Path("blog") / "views.py").read_text()
    assert "@login_required" in views
    assert "AuthenticatedHttpRequest" in views


def test_crud_config_pyproject_skip_git_check_set(django_project, runner: CommandRunner):
    create_pyproject_crud_config(skip_git_check=True)
    runner.invoke("crud", "blog.post")
    assert healthy_django_project()
    views = (Path("blog") / "views.py").read_text()
    assert "post_list" in views


def test_crud_config_pyproject_login_required(django_project, runner: CommandRunner):
    create_pyproject_crud_config(skip_git_check=True, login_required=True)
    runner.invoke("crud", "blog.post")
    assert healthy_django_project()
    views = (Path("blog") / "views.py").read_text()
    assert "@login_required" in views
    assert "AuthenticatedHttpRequest" in views


def test_crud_config_pyproject_blueprints(django_project, runner: CommandRunner):
    bp = django_project / "blueprints"
    bp.mkdir()
    html_file = bp / "dummy.html.jinja"
    html_file.touch()
    html_file.write_text("{{ model_name }}")
    create_pyproject_crud_config(blueprints=str(Path("blueprints")), skip_git_check=True)
    runner.invoke("crud", "blog.post")
    views = (Path("blog") / "views.py").read_text()
    rendered_file = Path("blog") / "templates" / "blog" / "post_dummy.html"
    assert rendered_file.exists()
    assert "Post" in rendered_file.read_text()
    assert "Post" in views


def test_crud_always_migrate(django_project, runner: CommandRunner, set_git_repo_to_clean):
    create_pyproject_crud_config(always_migrate=True)
    settings = django_project / "myproject" / "settings.py"
    settings.write_text(settings.read_text() + "\n\n" + "INSTALLED_APPS += ['non_existent_app']\n")
    with pytest.raises(cappa.Exit):
        runner.invoke("crud", "blog.post")
    assert "post_list" not in Path("blog/views.py").read_text()
    assert not healthy_django_project()


def test_forms_dates_widgets(django_project, runner: CommandRunner, set_git_repo_to_clean):
    runner.invoke("start-app", "authors")
    models = django_project / "myproject" / "authors" / "models.py"
    models.write_text(
        models.read_text()
        + "\n    birthday=models.DateField()\n"
        + "    tarted_writing=models.DateTimeField()\n"
        + "    wake_up_time=models.TimeField()\n"
    )
    runner.invoke("crud", "authors.author")
    runner.invoke("crud", "blog.post")
    assert healthy_django_project()
    assert "widgets" not in (django_project / "blog" / "forms.py").read_text()
    forms = django_project / "myproject" / "authors" / "forms.py"
    assert "widgets" in forms.read_text()
    assert "DateInput" in forms.read_text()
    assert "TimeInput" in forms.read_text()
    assert "DateTimeInput" in forms.read_text()



from pathlib import Path

from cappa.testing import CommandRunner
from tests.commands.test_crud import healthy_django_project


def test_start_app(django_project, runner: CommandRunner):
    runner.invoke("start-app", "products")
    assert Path("myproject/products").exists()
    assert healthy_django_project()



from django.core.management import call_command
from django.core.management.base import BaseCommand
from django.db import connection
from falco.management.base import exit_if_debug_false
from falco.management.base import get_apps_dir
from falco.management.commands.rm_migrations import Command as RmMigrationsCommand
from falco.utils import simple_progress


class Command(BaseCommand):
    help = "Delete and recreate all migrations while keeping the database data."
    requires_migrations_checks = True

    def handle(self, *_, **__):
        exit_if_debug_false()
        # TODO: should stop if all current migrations are not applied
        apps_dir = get_apps_dir()
        RmMigrationsCommand.delete_migration_files(apps_dir=apps_dir)
        with simple_progress("Resetting migrations..."):
            with connection.cursor() as cursor:
                cursor.execute("DELETE FROM django_migrations")
            call_command("makemigrations")
            call_command("migrate", "--fake")
        self.stdout.write(self.style.SUCCESS("Done!"))



# Generated by Django 5.1.2 on 2024-10-12 21:56
import django.db.models.deletion
from django.db import migrations
from django.db import models


class Migration(migrations.Migration):
    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="Author",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("updated_at", models.DateTimeField(auto_now=True)),
                ("name", models.CharField(max_length=255)),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="Book",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("updated_at", models.DateTimeField(auto_now=True)),
                ("name", models.CharField(max_length=255)),
                ("slug", models.SlugField()),
                ("description", models.TextField()),
                ("published_at", models.DateField()),
                ("on_going", models.BooleanField(default=True)),
                ("cover_art", models.FileField(upload_to="covers")),
                (
                    "author",
                    models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to="books.author"),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
    ]



from pathlib import Path

from falco.management.base import CleanRepoOnlyCommand
from falco.management.base import exit_if_debug_false
from falco.management.base import get_apps_dir
from falco.utils import simple_progress


class Command(CleanRepoOnlyCommand):
    help = "Remove all migrations for the specified applications directory, intended only for development."

    def handle(self, *_, **__):
        exit_if_debug_false()
        apps_dir = get_apps_dir()
        apps = self.delete_migration_files(apps_dir=apps_dir)
        self.stdout.write(self.style.SUCCESS(f"Removed migration files for apps: {', '.join(apps)}"))

    @classmethod
    def delete_migration_files(cls, apps_dir: Path) -> set[str]:
        apps = set()
        with simple_progress("Removing migration files"):
            for folder in apps_dir.iterdir():
                migration_dir = folder / "migrations"
                if not migration_dir.exists():
                    continue
                apps.add(folder.stem)
                for file in migration_dir.iterdir():
                    if file.suffix == ".py" and file.name not in ["__init__.py"]:
                        file.unlink()
        return apps



from django.conf import settings


class DBTaskRouter:
    """
    A router to control all database operations on models in the django_tasks application.
    """

    route_app_labels = {"django_tasks_database", "django_tasks", "django.tasks"}

    def db_for_read(self, model, **hints):
        """
        Attempts to read django_tasks models go to the tasks database.
        """
        if model._meta.app_label in self.route_app_labels:  # noqa
            return self.get_db()
        return None

    def db_for_write(self, model, **hints):
        """
        Attempts to write django_tasks models go to the tasks database.
        """
        if model._meta.app_label in self.route_app_labels:  # noqa
            return self.get_db()
        return None

    def allow_migrate(self, db, app_label, model_name=None, **hints):  # noqa
        """
        Make sure the django_tasks app only appears in the tasks database.
        """
        if app_label in self.route_app_labels:
            return db == self.get_db()
        return None

    @classmethod
    def get_db(cls):
        try:
            return settings.TASKS["default"]["OPTIONS"]["database"]
        except (AttributeError, KeyError):
            return "tasks_db"



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

from django.conf import settings
from django.contrib.staticfiles import finders
from django.http import FileResponse
from django.http import HttpResponse
from django.shortcuts import render
from django.utils import timezone
from django.views.decorators.cache import cache_control
from django.views.decorators.http import require_GET
from falco.conf import app_settings
from falco.decorators import login_not_required

if TYPE_CHECKING:
    from django.http import HttpRequest


@require_GET
@cache_control(
    max_age=0 if settings.DEBUG else app_settings.CACHE_TIME_ROBOTS_TXT,
    immutable=True,
    public=True,
)
@login_not_required
def robots_txt(request: HttpRequest) -> HttpResponse:
    return render(request, app_settings.TEMPLATE_ROBOTS_TXT, content_type="text/plain")


@require_GET
@cache_control(
    max_age=0 if settings.DEBUG else app_settings.CACHE_TIME_SECURITY_TXT,
    immutable=True,
    public=True,
)
@login_not_required
def security_txt(request: HttpRequest) -> HttpResponse:
    return render(
        request,
        app_settings.TEMPLATE_SECURITY_TXT,
        context={
            "year": timezone.now().year + 1,
        },
        content_type="text/plain",
    )


@require_GET
@cache_control(
    max_age=0 if settings.DEBUG else app_settings.CACHE_TIME_FAVICON,
    immutable=True,
    public=True,
)
@login_not_required
def favicon(request: HttpRequest) -> HttpResponse | FileResponse:
    name = request.path.lstrip("/")
    if path := finders.find(name):
        return FileResponse(Path(path).read_bytes())
    return HttpResponse(
        (
            '<svg xmlns="http://www.w3.org/2000/svg" viewBox="0 0 100 100">'
            '<text y=".9em" font-size="90">🚀</text>'
            "</svg>"
        ),
        content_type="image/svg+xml",
    )



# Generated by Django 5.1.2 on 2024-10-14 06:55

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("books", "0001_initial"),
    ]

    operations = [
        migrations.AlterField(
            model_name="book",
            name="cover_art",
            field=models.FileField(blank=True, upload_to="covers"),
        ),
    ]



# inspriation from https://noumenal.es/notes/tailwind/django-integration/
from __future__ import annotations

from pathlib import Path

from django.conf import settings
from django.core.management.base import BaseCommand
from django.template.utils import get_app_template_dirs


class Command(BaseCommand):
    help = "List all template files of the Django project."

    def handle(self, *_, **__):
        app_template_dirs = get_app_template_dirs("templates")
        setting_template_dirs = [Path(str(dir_)) for dir_ in settings.TEMPLATES[0]["DIRS"]]

        template_files = []
        for dir_ in [*app_template_dirs, *setting_template_dirs]:
            template_files.extend(self.list_template_files(dir_))

        self.stdout.write("\n".join(template_files))

    def list_template_files(self, template_dir: Path) -> list[str]:
        return [str(file_path) for file_path in Path(template_dir).rglob("*") if file_path.suffix in [".html", ".txt"]]



from django.urls import path

from . import views

app_name = "entries"

urlpatterns = [
    path("entries/", views.entry_list, name="entry_list"),
    path("entries/create/", views.entry_create, name="entry_create"),
    path("entries/<int:pk>/", views.entry_detail, name="entry_detail"),
    path("entries/<int:pk>/update/", views.entry_update, name="entry_update"),
    path("entries/<int:pk>/delete/", views.entry_delete, name="entry_delete"),
]



from __future__ import annotations

from importlib.util import find_spec
from typing import TYPE_CHECKING

from falco.conf import app_settings

if find_spec("sentry_sdk"):
    if TYPE_CHECKING:
        from sentry_sdk._types import SamplingContext

    def sentry_traces_sampler(sampling_context: SamplingContext):
        if _should_disregard(sampling_context):
            return 0

        return app_settings.SENTRY_TRACES_RATE

    def sentry_profiles_sampler(sampling_context: SamplingContext):
        if _should_disregard(sampling_context):
            return 0

        return app_settings.SENTRY_PROFILE_RATE

    def _should_disregard(sampling_context: SamplingContext) -> bool:
        DISGARDED_METHODS = app_settings.SENTRY_DISGARDED_METHODS  # noqa
        DISGARDED_PATHS = app_settings.SENTRY_DISGARDED_PATHS  # noqa

        return (
            sampling_context.get("wsgi_environ", None) is not None
            and sampling_context["wsgi_environ"]["REQUEST_METHOD"] in DISGARDED_METHODS
            and sampling_context["wsgi_environ"]["PATH_INFO"] in DISGARDED_PATHS
        )



from __future__ import annotations

from pathlib import Path

ALLOWED_HOSTS = ["*"]

CACHES = {
    "default": {
        "BACKEND": "django.core.cache.backends.dummy.DummyCache",
    }
}

DEBUG = False

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": ":memory:",
    }
}

EMAIL_BACKEND = "django.core.mail.backends.locmem.EmailBackend"

# INSTALLED_APPS = [
#     "django_twc_toolbox",
#     "django_twc_toolbox.crud",
#     "django.contrib.admin",
#     "django.contrib.auth",
#     "django.contrib.contenttypes",
#     # "django_tables2",
#     # "simple_history",
#     "template_partials",
#     # "tests.dummy",
#     # "tests.test_crud",
# ]

LOGGING_CONFIG = None

PASSWORD_HASHERS = [
    "django.contrib.auth.hashers.MD5PasswordHasher",
]

SECRET_KEY = "not-a-secret"

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [Path(__file__).parent / "templates"],
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.request",
            ],
            "loaders": [
                (
                    "template_partials.loader.Loader",
                    [
                        "django.template.loaders.filesystem.Loader",
                        "django.template.loaders.app_directories.Loader",
                    ],
                )
            ],
        },
    }
]



from django.urls import path
from django.views import defaults as default_views

from .views import favicon

favicon_urlpatterns = [
    path("android-chrome-192x192.png", favicon),
    path("android-chrome-512x512.png", favicon),
    path("apple-touch-icon.png", favicon),
    path("browserconfig.xml", favicon),
    path("favicon-16x16.png", favicon),
    path("favicon-32x32.png", favicon),
    path("favicon.ico", favicon),
    path("mstile-150x150.png", favicon),
    path("safari-pinned-tab.svg", favicon),
]


errors_urlpatterns = [
    path(
        "400/",
        default_views.bad_request,
        kwargs={"exception": Exception("Bad Request!")},
    ),
    path(
        "403/",
        default_views.permission_denied,
        kwargs={"exception": Exception("Permission Denied")},
    ),
    path(
        "404/",
        default_views.page_not_found,
        kwargs={"exception": Exception("Page not Found")},
    ),
    path("500/", default_views.server_error),
]



import importlib.util
import subprocess
from multiprocessing import Pool

from django.conf import settings
from django.core.management import call_command
from django.core.management.base import BaseCommand
from falco.conf import app_settings


class Command(BaseCommand):
    help = "Run every dev process needed in one command"

    def add_arguments(self, parser):
        parser.add_argument(
            "address",
            type=str,
            nargs="?",
            default="127.0.0.1:8000",
            help="Address to run the django server on",
        )
        parser.add_argument(
            "-p",
            "--plus",
            action="store_true",
            help="Run the server using runserver_plus",
        )

    def handle(self, *_, **options):
        address = options["address"]
        use_runserver_plus = options["plus"]
        commands = {
            "runserver": (
                "python -m django runserver_plus {address}"
                if use_runserver_plus
                else "python -m django runserver {address}"
            )
        }
        if "django_tailwind_cli" in settings.INSTALLED_APPS:
            commands["tailwind"] = f"python -m django tailwind watch"
        if "tailwind" in settings.INSTALLED_APPS:
            commands["tailwind"] = f"python -m django tailwind start"
        if "django_q" in settings.INSTALLED_APPS:
            commands["qcluster"] = f"python -m django qcluster"

        try:
            run_db_worker = settings.TASKS["default"]["BACKEND"] in [
                "django_tasks.backends.database.DatabaseBackend",
                "django.tasks.backends.database.DatabaseBackend",
            ]
        except (AttributeError, KeyError):
            run_db_worker = False

        if run_db_worker:
            commands["worker"] = f"python -m django db_worker -v 3"

        commands.update(app_settings.WORK)
        commands["runserver"] = commands["runserver"].format(address=address)

        call_command("migrate")
        if importlib.util.find_spec("honcho"):
            self.run_with_honcho(commands)
        else:
            self.run_with_multiprocess(commands)

    @classmethod
    def run_with_multiprocess(cls, commands: dict):
        with Pool(processes=len(commands)) as pool:
            try:
                pool.map(subprocess.run, [cmd.split() for cmd in commands.values()])
            except KeyboardInterrupt:
                pool.terminate()
            finally:
                pool.close()
                pool.join()

    @classmethod
    def run_with_honcho(cls, commands: dict):
        from honcho.manager import Manager

        manager = Manager()
        for name, cmd in commands.items():
            manager.add_process(name, cmd)
        try:
            manager.loop()
        finally:
            manager.terminate()



import importlib.util

from django.contrib.auth.models import User
from django.http import HttpRequest as HttpRequestBase

if importlib.util.find_spec("django_htmx"):
    # TODO: should look for the middleware, not the package
    from django_htmx.middleware import HtmxDetails

    class HttpRequest(HttpRequestBase):
        htmx: HtmxDetails

else:

    class HttpRequest(HttpRequestBase):
        pass


class AuthenticatedHttpRequest(HttpRequest):
    user: User



from functools import wraps

from django.http import HttpResponse
from django.template.loader import render_to_string
from falco.types import HttpRequest


def for_htmx(
    *,
    if_hx_target: str | None = None,
    use_template: str | None = None,
    use_partial: str | list[str] | None = None,
    use_partial_from_params: bool = False,
):
    """
    Adapted from https://github.com/spookylukey/django-htmx-patterns/blob/master/code/htmx_patterns/utils.py
    If the request is from htmx, then render a partial page, using either:
    - the template specified in `use_template` param
    - the partial/partials specified in `use_partial` param
    - the partial/partials specified in GET/POST parameter "use_partial", if `use_partial_from_params=True` is passed

    If the optional `if_hx_target` parameter is supplied, the
    hx-target header must match the supplied value as well in order
    for this decorator to be applied.
    """
    if len([p for p in [use_partial, use_template, use_partial_from_params] if p]) != 1:
        msg = "You must pass exactly one of 'use_template', 'use_partial' or 'use_partial_from_params=True'"
        raise ValueError(msg)

    def decorator(view):
        @wraps(view)
        def _view(request: HttpRequest, *args, **kwargs):
            resp = view(request, *args, **kwargs)

            if not request.htmx:
                return resp

            apply_decorator = if_hx_target is None or request.headers.get("Hx-Target", None) == if_hx_target
            if not apply_decorator:
                return resp

            partials_to_use = use_partial
            if not hasattr(resp, "render"):
                if not resp.content and any(
                    h in resp.headers
                    for h in (
                        "Hx-Trigger",
                        "Hx-Trigger-After-Swap",
                        "Hx-Trigger-After-Settle",
                        "Hx-Redirect",
                    )
                ):
                    # This is a special case response, it doesn't need modifying:
                    return resp

                msg = "Cannot modify a response that isn't a TemplateResponse"
                raise ValueError(msg)
            if resp.is_rendered:
                msg = "Cannot modify a response that has already been rendered"
                raise ValueError(msg)

            if use_partial_from_params:
                use_partial_from_params_val = _get_param_from_request(request, "use_partial")
                if use_partial_from_params_val is not None:
                    partials_to_use = use_partial_from_params_val

            if use_template is not None:
                resp.template_name = use_template
            elif partials_to_use is not None:
                if not isinstance(partials_to_use, list):
                    partials_to_use = [partials_to_use]

                rendered_partials = [
                    render_to_string(
                        f"{resp.template_name}#{b}",
                        context=resp.context_data,
                        request=request,
                    )
                    for b in partials_to_use
                ]
                # Create new simple HttpResponse as replacement
                resp = HttpResponse(
                    content="".join(rendered_partials),
                    status=resp.status_code,
                    headers=resp.headers,
                )

            return resp

        return _view

    return decorator


def _get_param_from_request(request, param):
    """
    Checks GET then POST params for specified param
    """
    if param in request.GET:
        return request.GET.getlist(param)
    if request.method == "POST" and param in request.POST:
        return request.POST.getlist(param)
    return None



from __future__ import annotations

import sys
from dataclasses import dataclass

from django.conf import settings

if sys.version_info >= (3, 12):
    from typing import override
else:  # pragma: no cover
    from typing_extensions import override  # pyright: ignore[reportUnreachable]


FALCO_SETTINGS_NAME = "FALCO"


@dataclass(frozen=True)
class AppSettings:
    CACHE_TIME_FAVICON = 60 * 60 * 24  # one day
    CACHE_TIME_ROBOTS_TXT = 60 * 60 * 24  # one day
    CACHE_TIME_SECURITY_TXT = 60 * 60 * 24  # one day
    TEMPLATE_ROBOTS_TXT = "robots.txt"
    TEMPLATE_SECURITY_TXT = ".well-known/security.txt"
    DEFAULT_PAGE_SIZE = 20
    SENTRY_DISGARDED_METHODS = ["GET", "HEAD"]
    SENTRY_DISGARDED_PATHS = ["/health/"]
    SENTRY_PROFILE_RATE = 0.5
    SENTRY_TRACES_RATE = 0.5
    WORK = {}

    @override
    def __getattribute__(self, __name: str) -> object:
        user_settings = getattr(settings, FALCO_SETTINGS_NAME, {})
        return user_settings.get(__name.lower(), super().__getattribute__(__name))  # pyright: ignore[reportAny]


app_settings = AppSettings()



from io import StringIO
from pathlib import Path
from typing import TypedDict

import parso
from django.apps import AppConfig
from django.apps import apps
from django.conf import settings
from django.core.management import call_command
from django.core.management.base import CommandError
from django.template import Context
from django.template import Template
from falco.management.base import CleanRepoOnlyCommand
from falco.management.commands.copy_template import get_template_absolute_path
from falco.utils import run_html_formatters
from falco.utils import run_python_formatters
from falco.utils import simple_progress

IMPORT_START_COMMENT = "# IMPORTS:START"
IMPORT_END_COMMENT = "# IMPORTS:END"
CODE_START_COMMENT = "# CODE:START"
CODE_END_COMMENT = "# CODE:END"


class DjangoField(TypedDict):
    verbose_name: str
    editable: bool
    class_name: str
    accessor: str  # e.g: {{product.name}}


class DjangoModel(TypedDict):
    name: str
    name_lower: str
    name_plural: str
    lookup_field: str
    path_converter: str
    verbose_name: str
    verbose_name_plural: str
    fields: dict[str, DjangoField]
    obj_accessor: str
    has_file_field: bool
    has_editable_date_field: bool


class BlueprintContext(TypedDict):
    login_required: bool
    entry_point: bool
    app_label: str
    model: DjangoModel
    fields_tuple: str
    editable_fields_tuple: str
    view_name_prefix: str
    list_view_name: str
    detail_view_name: str
    delete_view_name: str
    list_view_url: str
    create_view_url: str
    detail_view_url: str
    update_view_url: str
    delete_view_url: str
    table_block: str
    pagination_block: str


class Command(CleanRepoOnlyCommand):
    help = "Generate CRUD (Create, Read, Update, Delete) views for a model."

    def add_arguments(self, parser):
        parser.add_argument(
            "model_path",
            type=str,
            help="The path (<app_label>.<model_name>) of the model to generate CRUD views for. Ex: myapp.product",
        )
        parser.add_argument(
            "-e",
            "--exclude",
            action="append",
            help="Fields to exclude from the views, forms and templates.",
        )
        parser.add_argument("--only-python", action="store_true", help="Generate only python code.")
        parser.add_argument("--only-html", action="store_true", help="Generate only html code.")
        parser.add_argument(
            "--entry-point",
            action="store_true",
            help="Use the specified model as the entry point of the app.",
        )
        parser.add_argument(
            "--login-required",
            "-l",
            action="store_true",
            help="Add the login_required decorator to all views.",
        )
        parser.add_argument(
            "--migrate",
            "-m",
            action="store_true",
            help="Run makemigrations and migrate beforehand",
        )

    def handle(self, *_, **options):
        model_path = options["model_path"]
        excluded_fields = options["exclude"] or []
        only_python = options["only_python"]
        only_html = options["only_html"]
        entry_point = options["entry_point"]
        login_required = options["login_required"]
        migrate = options["migrate"]

        model_path_parts = model_path.split(".")
        if len(model_path_parts) == 1:
            model_name = None
            app_label = model_path_parts[0]
        else:
            model_name = model_path_parts.pop()
            app_label = ".".join(model_path_parts)

        if entry_point and not model_name:
            msg = "The --entry-point option requires a full model path."
            raise CommandError(msg)

        if migrate:
            with simple_progress("Running migrations"):
                call_command("makemigrations", app_label)
                call_command("migrate")

        app: AppConfig = apps.get_app_config(app_label)
        with simple_progress("Getting models info"):
            all_django_models = self.get_models_data(
                app_label=app_label,
                excluded_fields=excluded_fields,
                entry_point=entry_point,
            )
            dirs = settings.TEMPLATES[0].get("DIRS", [])
            templates_dir = Path(dirs[0]) / app_label if dirs else Path(app.path) / "templates"

        django_models = (
            [m for m in all_django_models if m["name"].lower() == model_name.lower()]
            if model_name
            else all_django_models
        )
        if model_name and not django_models:
            msg = f"Model {model_name} not found in app {app_label}"
            raise CommandError(msg)

        blueprint_contexts: list[BlueprintContext] = [
            build_blueprint_context(
                app_label=app_label,
                django_model=django_model,
                login_required=login_required,
                entry_point=entry_point,
            )
            for django_model in django_models
        ]

        updated_python_files = set()

        if not only_html:
            python_blueprints = ("crud/forms.py.dtl", "crud/views.py.dtl")
            updated_python_files.update(
                self.generate_python_code(
                    app=app,
                    blueprints=[Path(get_template_absolute_path(p)) for p in python_blueprints],
                    contexts=blueprint_contexts,
                )
            )

            updated_python_files.update(
                self.generating_urls(
                    app=app,
                    django_models=django_models,
                    entry_point=entry_point,
                )
            )

        updated_html_files = set()
        if not only_python:
            html_blueprints = ("crud/list.html", "crud/form.html", "crud/detail.html")
            updated_html_files.update(
                self.generate_html_templates(
                    contexts=blueprint_contexts,
                    entry_point=entry_point,
                    blueprints=[Path(get_template_absolute_path(p)) for p in html_blueprints],
                    templates_dir=templates_dir,
                )
            )

        with simple_progress("Running python formatters"):
            for file in updated_python_files:
                run_python_formatters(str(file))

        with simple_progress("Running html formatters"):
            for file in updated_html_files:
                run_html_formatters(str(file))

        display_names = ", ".join(m.get("name") for m in django_models)
        self.stdout.write(self.style.SUCCESS(f"CRUD views generated for: {display_names}"))

    @classmethod
    def get_models_data(cls, app_label: str, excluded_fields: list[str], *, entry_point: bool) -> "list[DjangoModel]":
        models = apps.get_app_config(app_label).get_models()
        file_fields = ("ImageField", "FileField")
        dates_fields = ("DateField", "DateTimeField", "TimeField")

        def get_model_dict(model) -> "DjangoModel":
            name = model.__name__
            name_lower = name.lower()
            if entry_point:
                name_plural = app_label.lower()
            else:
                name_plural = f"{name.replace('y', 'ies')}" if name.endswith("y") else f"{name}s"

            verbose_name = model._meta.verbose_name  # noqa
            verbose_name_plural = model._meta.verbose_name_plural  # noqa

            def get_accessor(field) -> str:
                base_accessor = f"{name_lower}.{field.name}"
                if field.__class__.__name__ in file_fields:
                    return f"{{% if {base_accessor} %}} {{{{ {base_accessor}.url }}}} {{% endif %}}"
                else:
                    return "{{" + base_accessor + "}}"

            fields: dict[str, DjangoField] = {
                field.name: {
                    "verbose_name": field.verbose_name,
                    "editable": field.editable,
                    "class_name": field.__class__.__name__,
                    "accessor": get_accessor(field)
                }
                for field in model._meta.fields  # noqa
                if field.name not in excluded_fields
            }
            name_lower = name.lower()
            lookup_field = getattr(model, "lookup_field", "pk")
            return {
                "name": name,
                "name_lower": name_lower,
                "name_plural": name_plural,
                "lookup_field": lookup_field,
                "path_converter": "int" if lookup_field == "pk" else "str",
                "fields": fields,
                "obj_accessor": "{{" + name_lower + "}}",
                "verbose_name": verbose_name,
                "verbose_name_plural": verbose_name_plural,
                "has_file_field": any(f["class_name"] in file_fields for f in fields.values()),
                "has_editable_date_field": any(
                    f["class_name"] in dates_fields and f["editable"] for f in fields.values()
                ),
            }

        return [get_model_dict(model) for model in models]

    @simple_progress("Generating python code")
    def generate_python_code(
        self,
        app: AppConfig,
        blueprints: list[Path],
        contexts: list["BlueprintContext"],
    ) -> list[Path]:
        updated_files = []

        for blueprint in blueprints:
            file_content = blueprint.read_text()
            imports_template = extract_content_from(file_content, IMPORT_START_COMMENT, IMPORT_END_COMMENT)
            code_template = extract_content_from(file_content, CODE_START_COMMENT, CODE_END_COMMENT)
            # blueprints python files end in .py.dtl
            file_name_without_jinja = ".".join(blueprint.name.split(".")[:-1])
            file_to_write_to = Path(app.path) / file_name_without_jinja
            file_to_write_to.touch(exist_ok=True)

            imports_content, code_content = "", ""

            for context in contexts:
                imports_content += render_from_string(imports_template, context)
                code_content += render_from_string(code_template, context)

            file_to_write_to.write_text(imports_content + file_to_write_to.read_text() + code_content)
            updated_files.append(file_to_write_to)

        model_name = contexts[0]["model"]["name"] if len(contexts) == 1 else None
        updated_files.append(self.register_models_in_admin(app=app, model_name=model_name))
        return updated_files

    @simple_progress("Generating urls")
    def generating_urls(
        self,
        app: AppConfig,
        django_models: list["DjangoModel"],
        *,
        entry_point: bool,
    ) -> list[Path]:
        urls_content = ""
        for django_model in django_models:
            model_name_lower = django_model["name_lower"]
            urlsafe_model_verbose_name_plural = django_model["verbose_name_plural"].lower().replace(" ", "-")
            view_name_prefix = "" if entry_point else f"{model_name_lower}_"
            list_view_name = "index" if entry_point else f"{model_name_lower}_list"
            prefix = "" if entry_point else f"{urlsafe_model_verbose_name_plural}/"
            urls_content += f"""
                path('{prefix}', views.{list_view_name}, name='{list_view_name}'),
                path('{prefix}new/', views.process_{view_name_prefix}form, name='{view_name_prefix}create'),
                path('{prefix}<{django_model['path_converter']}:{django_model['lookup_field']}>/', views.{view_name_prefix}detail, name='{view_name_prefix}detail'),
                path('{prefix}<{django_model['path_converter']}:{django_model['lookup_field']}>/edit/', views.process_{view_name_prefix}form, name='{view_name_prefix}update'),
                path('{prefix}<{django_model['path_converter']}:{django_model['lookup_field']}>/delete/', views.{view_name_prefix}delete, name='{view_name_prefix}delete'),
            """

        app_urls = Path(app.path) / "urls.py"
        updated_files = [app_urls]
        if app_urls.exists() and app_urls.read_text().strip() != "":
            urlpatterns = f"\nurlpatterns +=[{urls_content}]"
            app_urls.write_text(app_urls.read_text() + urlpatterns)
        else:
            app_urls.touch()
            app_urls.write_text(
                f"""
from django.urls import path
from . import views

app_name = "{app.label}"

urlpatterns = [
{urls_content}
]
        """
            )
            root_url = settings.ROOT_URLCONF
            root_url = root_url.strip().replace(".", "/")
            root_url_path = Path(f"{root_url}.py")
            module = parso.parse(root_url_path.read_text())
            new_path = parso.parse(f"path('{app.label}/', include('{app.name}.urls', namespace='{app.label}'))")
            for node in module.children:
                try:
                    if (
                        node.children[0].type == parso.python.tree.ExprStmt.type
                        and node.children[0].children[0].value == "urlpatterns"
                    ):
                        patterns = node.children[0].children[2]
                        elements = patterns.children[1]
                        elements.children.append(new_path)
                        new_content = module.get_code()
                        new_content = "from django.urls import include\n" + new_content
                        root_url_path.write_text(new_content)
                        break
                except AttributeError:
                    continue
            updated_files.append(root_url_path)
        return updated_files

    @simple_progress("Generating html templates")
    def generate_html_templates(
        self,
        templates_dir: Path,
        blueprints: list[Path],
        contexts: list["BlueprintContext"],
        *,
        entry_point: bool,
    ) -> list[Path]:
        updated_files = []
        templates_dir.mkdir(exist_ok=True, parents=True)
        for blueprint in blueprints:
            for context in contexts:
                model_name_lower = context["model"]["name_lower"]
                if entry_point:
                    new_filename = "index.html" if blueprint.name == "list.html" else blueprint.name
                else:
                    new_filename = f"{model_name_lower}_{blueprint.name}"
                file_to_write_to = templates_dir / new_filename
                file_to_write_to.touch(exist_ok=True)
                views_content = render_from_string(blueprint.read_text(), context=context)

                file_to_write_to.write_text(views_content)
                updated_files.append(file_to_write_to)

        return updated_files

    def register_models_in_admin(self, app: AppConfig, model_name: str | None = None) -> Path:
        admin_file = Path(app.path) / "admin.py"

        # Skip further processing if model_name is not specified and file is non-empty
        if not model_name and admin_file.exists() and admin_file.stat().st_size > 0:
            self.stdout.write(self.style.WARNING("Skipping admin registration as the file is not empty."))
            return admin_file

        admin_file.touch(exist_ok=True)
        cmd_args = [app.label]
        if model_name:
            cmd_args.append(model_name)

        try:
            output = StringIO()
            call_command("admin_generator", *cmd_args, stdout=output)
        except CommandError as e:
            self.stdout.write(self.style.WARNING(f"Admin failed to generate: {e}"))
            return admin_file

        # the first line of the generated code set the encoding, it is useless for python 3
        admin_code = output.getvalue().split("\n", 1)[1]
        existing_code = admin_file.read_text()

        if model_name and model_name.title() in existing_code:
            self.stdout.write(self.style.WARNING(f"Model {model_name} is already registered."))
            return admin_file

        admin_file.write_text(existing_code + admin_code)

        if not model_name:
            # we probably don't need to reorder the imports if the admin code is being generated for all models
            return admin_file

        # if this is not the first time running this, the imports will be messed up, move all
        # of them to the top
        admin_lines = admin_file.read_text().split("\n")
        _imports = []
        _code = []
        for line in admin_lines:
            if line.startswith("from"):
                _imports.append(line)
            else:
                _code.append(line)
        admin_file.write_text("\n" + "\n".join(_imports) + "\n" + "\n".join(_code))

        return admin_file


def build_blueprint_context(
    app_label: str,
    django_model: DjangoModel,
    *,
    entry_point: bool,
    login_required: bool,
) -> BlueprintContext:
    model_name_lower = django_model["name_lower"]
    view_name_prefix = "" if entry_point else f"{model_name_lower}_"
    list_view_name = "index" if entry_point else f"{model_name_lower}_list"
    detail_view_name = f"{view_name_prefix}detail"
    update_view_name = f"{view_name_prefix}update"
    delete_view_name = f"{view_name_prefix}delete"
    create_view_url = f"{{% url '{app_label}:{view_name_prefix}create' %}}"
    return {
        "app_label": app_label,
        "model": django_model,
        "fields_tuple": tuple(django_model["fields"].keys()),
        "editable_fields_tuple": tuple(key for key, value in django_model["fields"].items() if value["editable"]),
        "view_name_prefix": view_name_prefix,
        "list_view_name": list_view_name,
        "detail_view_name": detail_view_name,
        "delete_view_name": delete_view_name,
        "list_view_url": f"{{% url '{app_label}:{list_view_name}' %}}",
        "create_view_url": create_view_url,
        "detail_view_url": f"{{% url '{app_label}:{detail_view_name}' {model_name_lower}.{django_model['lookup_field']} %}}",
        "update_view_url": f"{{% url '{app_label}:{update_view_name}' {model_name_lower}.{django_model['lookup_field']} %}}",
        "delete_view_url": f"{{% url '{app_label}:{delete_view_name}' {model_name_lower}.{django_model['lookup_field']} %}}",
        "entry_point": entry_point,
        "login_required": login_required,
        "pagination_block": f"""
        {{% if {model_name_lower}s_page.paginator.num_pages > 1 %}}
            {{% include "partials/pagination.html" with page={model_name_lower}s_page %}}
        {{% endif %}}
        """,
        "table_block": f"""
        {{% if {model_name_lower}s_page.object_list %}}
        {{% include "partials/table.html" with objects={model_name_lower}s_page.object_list fields=fields detail_view="{app_label}:{detail_view_name}" delete_view="{app_label}:{delete_view_name}" update_view="{app_label}:{update_view_name}" %}}
        {{% else %}}
        <p class="mt-8">There are no {django_model["verbose_name_plural"]}. <a class="hover:underline cursor-pointer" href="{create_view_url}">Create one now?</a> </p>
        {{% endif %}}
        """,
    }


def render_from_string(template_string: str, context: dict) -> str:
    return Template(template_string).render(Context(context))


def extract_content_from(text: str, start_comment: str, end_comment: str):
    start_index = text.find(start_comment) + len(start_comment)
    end_index = text.find(end_comment)
    return text[start_index:end_index]



from __future__ import annotations

import pytest
from django.contrib.auth import get_user_model
from django.contrib.auth.models import AnonymousUser
from django.core.paginator import Paginator
from django.http import QueryDict
from django.template import Context
from django.template import RequestContext
from django.template import Template
from model_bakery import baker

from django_twc_toolbox.templatetags.django_twc_toolbox import class_name
from django_twc_toolbox.templatetags.django_twc_toolbox import display_name
from django_twc_toolbox.templatetags.django_twc_toolbox import elided_page_range
from falco.templatetags.django_twc_toolbox import initials
from django_twc_toolbox.templatetags.django_twc_toolbox import klass
from django_twc_toolbox.templatetags.django_twc_toolbox import query_string
from django_twc_toolbox.templatetags.django_twc_toolbox import startswith

pytestmark = pytest.mark.django_db

User = get_user_model()


@pytest.mark.parametrize(
    "model,kwargs,expected",
    [
        (User, {"username": "johndoe", "first_name": "John", "last_name": "Doe"}, "JD"),
        (User, {"username": "janedoe"}, "J"),
        (AnonymousUser, {}, "N/A"),
    ],
)
def test_initials(model, kwargs, expected):
    user = model(**kwargs)

    assert initials(user) == expected


@pytest.mark.parametrize(
    "model,kwargs,expected",
    [
        (User, {"username": "johndoe", "first_name": "John", "last_name": "Doe"}, "JD"),
        (User, {"username": "janedoe"}, "J"),
        (AnonymousUser, {}, "N/A"),
    ],
)
def test_initials_templatet_ag(model, kwargs, expected):
    if kwargs:
        user = baker.make(model, **kwargs)
    else:
        user = model()

    template = Template("{% load django_twc_toolbox %} Initials: {{ user|initials }}")

    rendered = template.render(Context({"user": user}))

    assert f"Initials: {expected}" in rendered


@pytest.mark.parametrize(
    "model,kwargs,expected",
    [
        (
            User,
            {"username": "johndoe", "first_name": "John", "last_name": "Doe"},
            "John Doe",
        ),
        (User, {"username": "janedoe", "first_name": "Jane"}, "Jane"),
        (User, {"username": "bobsmith"}, "Bobsmith"),
        (AnonymousUser, {}, "N/A"),
    ],
)
def test_display_name(model, kwargs, expected):
    user = model(**kwargs)

    assert display_name(user) == expected


@pytest.mark.parametrize(
    "model,kwargs,expected",
    [
        (
            User,
            {"username": "johndoe", "first_name": "John", "last_name": "Doe"},
            "John Doe",
        ),
        (User, {"username": "janedoe", "first_name": "Jane"}, "Jane"),
        (User, {"username": "bobsmith"}, "Bobsmith"),
        (AnonymousUser, {}, "N/A"),
    ],
)
def test_display_name_template_tag(model, kwargs, expected):
    if kwargs:
        user = baker.make(model, **kwargs)
    else:
        user = model()

    template = Template("{% load django_twc_toolbox %} Name: {{ user|display_name }}")

    rendered = template.render(Context({"user": user}))

    assert f"Name: {expected}" in rendered










from pathlib import Path

import parso
from django.conf import settings
from django.core.management import CommandError
from django.core.management import call_command
from django.core.management.base import BaseCommand
from falco.management.base import get_apps_dir
from falco.utils import run_python_formatters
from falco.utils import simple_progress


class Command(BaseCommand):
    help = "Initialize a new django app the falco way."

    def add_arguments(self, parser):
        parser.add_argument("app_name", type=str, help="Name of the app to create.")

    def handle(self, *_, **options):
        app_name = options["app_name"]
        apps_dir = get_apps_dir()
        app_dir = apps_dir / app_name
        final_app_name = f"{apps_dir.name}.{app_name}".strip(".").lower()

        try:
            app_dir.mkdir()
        except FileExistsError as e:
            msg = f"Python module with the name {app_name} already exists in {apps_dir}."
            raise CommandError(msg) from e

        self.create_app(app_name=app_name, app_dir=app_dir, final_app_name=final_app_name)
        self.register_app(app_name=final_app_name)

    @classmethod
    def create_app(cls, app_name: str, app_dir: Path | None, final_app_name: str):
        with simple_progress(f"Creating {app_name} app"):
            call_command("startapp", app_name, str(app_dir))

            (app_dir / "tests.py").unlink()

            model_name = app_name[:-1].capitalize() if app_name.endswith("s") else app_name.capitalize()

            models_file = app_dir / "models.py"

            models_file.write_text(
                f"""
from django.db import models
from falco.models import TimeStamped

class {model_name}(TimeStamped):
    name=models.CharField(max_length=255)

    def __str__(self):
        return self.name
"""
            )

            (app_dir / "admin.py").write_text("")

            (app_dir / "views.py").write_text("")

            app_config_file = app_dir / "apps.py"
            app_config_file.write_text(app_config_file.read_text().replace(app_name, final_app_name))

        run_python_formatters(models_file)

    @simple_progress("Registering app")
    def register_app(self, app_name: str):
        names = ["LOCAL_APPS", "INSTALLED_APPS"]

        settings_file = Path(settings.SETTINGS_MODULE.replace(".", "/") + ".py")

        module = parso.parse(settings_file.read_text())

        for node in module.children:
            try:
                if (
                    node.children[0].type == parso.python.tree.ExprStmt.type
                    and node.children[0].children[0].value in names
                ):
                    apps = node.children[0].children[2]
                    elements = apps.children[1]

                    elements.children.append(parso.parse(f"'{app_name}',"))
                    new_content = module.get_code()
                    settings_file.write_text(new_content)
                    break
            except AttributeError:
                continue
        run_python_formatters(settings_file)



from pathlib import Path
from unittest.mock import patch

import pytest
from cappa.testing import CommandRunner
from falco.management.commands.htmx import Htmx
from falco_cli.config import read_falco_config
from falco_cli.config import write_falco_config


@pytest.fixture(autouse=True)
def mock_latest_tag_getter():
    def _get_latest_tag():
        return "1.9.10"

    with patch("falco.management.commands.htmx.get_latest_tag", new=_get_latest_tag):
        yield


def test_htmx_download(runner: CommandRunner):
    runner.invoke("htmx")
    assert Path("htmx.min.js").exists()


def test_htmx_download_with_version(runner: CommandRunner):
    runner.invoke("htmx", "latest")
    assert Path("htmx.min.js").exists()


def test_htmx_download_with_specific_version(runner: CommandRunner):
    runner.invoke("htmx", "1.8.0")
    assert Path("htmx.min.js").exists()


def test_htmx_download_to_output_dir(runner: CommandRunner):
    output = Path("htmx/vendors")
    runner.invoke("htmx", "-o", str(output.resolve()))
    assert (output / "htmx.min.js").exists()


def test_htmx_download_to_output_file(runner: CommandRunner):
    output = Path("htmx/vendors/htmx.js")
    runner.invoke("htmx", "-o", str(output.resolve()))
    assert output.exists()


def test_htmx_with_pyproject_toml(runner: CommandRunner):
    pyproject_toml = Path("pyproject.toml")
    pyproject_toml.touch()
    write_falco_config(pyproject_path=pyproject_toml)
    runner.invoke("htmx")
    assert Path("htmx.min.js").exists()
    filepath, version = Htmx.read_from_config(read_falco_config(pyproject_toml))
    assert filepath == Path("htmx.min.js")


def test_htmx_with_pyproject_toml_custom_folder(runner: CommandRunner):
    pyproject_toml = Path("pyproject.toml")
    pyproject_toml.touch()
    write_falco_config(pyproject_path=pyproject_toml)
    runner.invoke("htmx", "-o", "static/htmx")
    output = Path("static/htmx/htmx.min.js")
    assert output.exists()
    filepath, version = Htmx.read_from_config(read_falco_config(pyproject_toml))
    assert filepath == output


def test_htmx_with_pyproject_toml_custom_file(runner: CommandRunner):
    pyproject_toml = Path("pyproject.toml")
    pyproject_toml.touch()
    write_falco_config(pyproject_path=pyproject_toml)
    runner.invoke("htmx", "-o", "static/htmx/htmx.js")
    output = Path("static/htmx/htmx.js")
    assert output.exists()
    filepath, version = Htmx.read_from_config(read_falco_config(pyproject_toml))
    assert filepath == output


def test_htmx_with_pyproject_toml_custom_file_existing_config(runner: CommandRunner):
    pyproject_toml = Path("pyproject.toml")
    pyproject_toml.touch()
    write_falco_config(pyproject_path=pyproject_toml, htmx="config/htmx/htmx.js")
    existing_path = Path("config/htmx/htmx.js")
    runner.invoke("htmx")
    filepath, _ = Htmx.read_from_config(read_falco_config(pyproject_toml))
    assert filepath == existing_path
    assert existing_path.exists()



from pathlib import Path

from cappa.testing import CommandRunner
from falco_cli.config import write_falco_config


def test_htmx_ext_download(runner: CommandRunner):
    runner.invoke("htmx-ext", "sse")
    assert Path("sse.js").exists()


# def test_htmx_ext_list_extensions(runner: CommandRunner):
#     result = runner.invoke("htmx-ext")
#     assert result.strip() != ""
#     assert "sse" in result


def test_htmx_ext_download_to_output_dir(runner: CommandRunner):
    output = Path("htmx/vendors/extensions")
    runner.invoke("htmx-ext", "sse", "-o", str(output.resolve()))
    assert (output / "sse.js").exists()


def test_htmx_ext_download_to_output_file(runner: CommandRunner):
    output = Path("htmx/vendors/extensions/sse.js")
    runner.invoke("htmx-ext", "sse", "-o", str(output.resolve()))
    assert output.exists()


def test_htmx_ext_file_existing_config(runner: CommandRunner):
    pyproject_toml = Path("pyproject.toml")
    pyproject_toml.touch()
    write_falco_config(pyproject_path=pyproject_toml, htmx="config/htmx/htmx.js")
    output = Path("config/htmx/sse.js")
    runner.invoke("htmx-ext", "sse")
    assert output.exists()


def test_htmx_ext_download_to_output_file_existing_config(runner: CommandRunner):
    pyproject_toml = Path("pyproject.toml")
    pyproject_toml.touch()
    write_falco_config(pyproject_path=pyproject_toml, htmx="config/htmx/htmx.js")
    output = Path("config/htmx/sse.js")
    runner.invoke("htmx-ext", "sse", "-o", ".")
    assert not output.exists()
    assert Path("sse.js").exists()



from __future__ import annotations

from contextlib import contextmanager
from pathlib import Path

import httpx
from django.conf import settings
from django.core.management.base import BaseCommand
from django.core.management.base import CommandError
from falco.utils import simple_progress
from httpx import codes

HTMX_DOWNLOAD_URL = "https://unpkg.com/htmx.org@{version}/dist/htmx.js"
HTMX_GH_RELEASE_LATEST_URL = "https://api.github.com/repos/bigskysoftware/htmx/releases/latest"


class Command(BaseCommand):
    help = "Download the latest version (if no version is specified) of htmx."

    def add_arguments(self, parser):
        parser.add_argument("version", type=str, default="latest")
        parser.add_argument(
            "-o",
            "--output-dir",
            type=Path,
            default=None,
            help="The directory to write the downloaded file to.",
        )

    def handle(self, *_, **options):
        version = options["version"]
        version = version if version != "latest" else get_latest_tag()
        output_dir = options.get("output_dir") or default_htmx_output_folder()
        url = HTMX_DOWNLOAD_URL.format(version=version)

        with network_request_with_progress(url, f"Downloading htmx version {version}") as response:
            content = response.content.decode("utf-8")
            if response.status_code == codes.NOT_FOUND:
                msg = f"Could not find htmx version {version}."
                raise CommandError(msg)

        filepath = output_dir / "htmx.js"
        filepath.write_text(f"/*htmx@{version}*/{content}")
        self.stdout.write(
            self.style.SUCCESS(
                f"htmx version {version} downloaded successfully to {filepath}",
            )
        )


def get_latest_tag() -> str:
    with network_request_with_progress(HTMX_GH_RELEASE_LATEST_URL, "Getting htmx latest version") as response:
        try:
            return response.json()["tag_name"][1:]
        except KeyError as e:
            msg = (
                "Unable to retrieve the latest version of htmx. "
                "This issue may be due to reaching the GitHub API rate limit. Please try again later."
            )
            raise CommandError(msg) from e


def default_htmx_output_folder() -> Path:
    try:
        folder = Path(settings.STATICFILES_DIRS[0]) / "vendors" / "htmx"
    except IndexError:
        msg = "Add at least one folder in your STATICFILES_DIRS settings and make sure it exists"
        raise CommandError(msg)
    folder.mkdir(exist_ok=True, parents=True)
    return folder


@contextmanager
def network_request_with_progress(url: str, description: str):
    try:
        with simple_progress(description):
            yield httpx.get(url)
    except httpx.ConnectError as e:
        msg = f"Connection error, {url} is not reachable."
        raise CommandError(msg) from e



from __future__ import annotations

import pytest
from django.apps import apps
from django.core.management import call_command
from django.core.management import CommandError
from django.test import override_settings
from falco.management.commands.copy_template import Command
from falco.management.commands.copy_template import get_template_absolute_path


@pytest.fixture
def command():
    return Command()


@pytest.fixture
def template_dirs(tmp_path):
    template_dir = tmp_path / "templates"
    template_dir.mkdir()

    source_dir = tmp_path / "source_templates"
    source_dir.mkdir()

    with override_settings(
        TEMPLATES=[
            {
                "BACKEND": "django.template.backends.django.DjangoTemplates",
                "DIRS": [str(template_dir), str(source_dir)],
            }
        ]
    ):
        yield template_dir, source_dir


@pytest.fixture
def create_template(template_dirs):
    def _create_template(name, content):
        _, source_dir = template_dirs
        template_path = source_dir / name
        template_path.write_text(content)
        return template_path

    return _create_template


def test_get_template_absolute_path(create_template):
    template_path = create_template("test_template.html", "<h1>Test Template</h1>")

    result = get_template_absolute_path("test_template.html")

    assert result == str(template_path)


def test_get_template_absolute_path_not_found():
    with pytest.raises(CommandError):
        get_template_absolute_path("non_existent_template.html")

    # assert result is None


def test_copy_template_with_destination(tmp_path, create_template):
    create_template("source_template.html", "<p>Source Template</p>")

    app_config = apps.get_app_config("books")

    app_dir = tmp_path / app_config.label
    app_dir.mkdir()

    with override_settings(BASE_DIR=tmp_path):
        original_path = app_config.path
        app_config.path = str(app_dir)

        call_command("copy_template", "source_template.html", "books")

        app_config.path = original_path

    copied_template = app_dir / "templates" / "source_template.html"

    assert copied_template.exists()
    assert copied_template.read_text() == "<p>Source Template</p>"


def test_copy_template_without_destination(tmp_path, create_template):
    create_template("source_template.html", "<p>Source Template</p>")

    with override_settings(BASE_DIR=tmp_path):
        call_command("copy_template", "source_template.html")

    copied_template = tmp_path / "templates" / "source_template.html"

    assert copied_template.exists()
    assert copied_template.read_text() == "<p>Source Template</p>"


def test_copy_template_without_destination_and_templates_setting(tmp_path, create_template):
    create_template("source_template.html", "<p>Source Template</p>")

    with override_settings(BASE_DIR=tmp_path, TEMPLATES=[]):
        # out = StringIO()
        with pytest.raises(CommandError):
            call_command("copy_template", "source_template.html")
    #     output = out.getvalue().strip()
    #
    # copied_template = tmp_path / "templates" / "source_template.html"
    #
    # assert output == "Source Template doesn't exist"
    # assert not copied_template.exists()


def test_copy_template_source_not_found(template_dirs):
    # out = StringIO()
    with pytest.raises(CommandError):
        call_command("copy_template", "nonexistent_template.html")
    # output = out.getvalue().strip()
    #
    # template_dir, _ = template_dirs
    # copied_template = template_dir / "nonexistent_template.html"
    #
    # assert output == "Source Template doesn't exist"
    # assert not copied_template.exists()



import subprocess

from cappa.testing import CommandRunner
from falco_cli.utils import run_in_shell


def makemigrations():
    subprocess.run(["python", "manage.py", "makemigrations"], check=False)


def migrate():
    subprocess.run(["python", "manage.py", "migrate"], check=False)


def change_model_attribute(django_project_dir):
    models_file = django_project_dir / "blog" / "models.py"
    models_file_content = models_file.read_text()
    models_file.write_text(
        models_file_content + "\n" + "    new_attribute = models.CharField(max_length=200, default='v')\n"
    )


def insert_a_post():
    from blog.models import Post

    Post.objects.create(title="t", content="c")


def count_nbr_of_posts() -> int:
    from blog.models import Post

    return Post.objects.all().count()


def count_migrations(django_project_dir):
    migrations_folder = django_project_dir / "blog" / "migrations"
    return len([file for file in migrations_folder.iterdir() if file.name.startswith("000")])


def test_reset_migrations(django_project, runner: CommandRunner, set_git_repo_to_clean):
    makemigrations()
    migrate()

    run_in_shell(insert_a_post, eval_result=False)
    count = run_in_shell(count_nbr_of_posts, eval_result=True)
    assert count == 1
    assert count == 1

    change_model_attribute(django_project)
    makemigrations()
    migrate()

    assert count_migrations(django_project) == 2
    runner.invoke("reset-migrations", ".")
    assert count_migrations(django_project) == 1

    count = run_in_shell(count_nbr_of_posts, eval_result=True)
    assert count == 1
